/********************************************************

	Copyright 2003 Masanao Izumo <iz@onicos.co.jp>

*********************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include "resample2.h"

#define RESAMPLE_EXTRA_LEN 256
#define PRECALC_LOOP_COUNT(start, end, incr) (((end) - (start) + (incr) - 1) / (incr))

#ifndef M_PI
#define M_PI		3.14159265358979323846	/* pi */
#endif

int
resample_init(Resampler *rs, int nch, int inbuf_max,
	      int src_freq, int dst_freq)
{
  int n;

  if (!inbuf_max)
    inbuf_max = RESAMPLE_DEFAULT_BUFSIZ;

  /* check */
  n = inbuf_max + RESAMPLE_EXTRA_LEN;
  n <<= RESAMPLE_FRAC_BITS;
  n >>= RESAMPLE_FRAC_BITS;
  if (n != inbuf_max + RESAMPLE_EXTRA_LEN) {
    fprintf(stderr, "Too large inbuf size: %d\n", inbuf_max);
    return -1;
  }

  memset(rs, 0, sizeof(*rs));
  rs->nch = nch;
  rs->src_freq = src_freq;
  rs->dst_freq = dst_freq;
  rs->inbuf_max = inbuf_max;
  rs->inbuf = (short *)malloc(nch * n * sizeof(short));
  rs->inbuf_size = n;
  rs->inbuf_len = 0;
  rs->ofs = 0;

  rs->f_incr = (double)src_freq / (double)dst_freq;
  rs->precalc_loop_gard = dst_freq / src_freq + 1;
  rs->incr = (int)(rs->f_incr * (1<<RESAMPLE_FRAC_BITS) + 0.5);
  if ((rs->incr>>RESAMPLE_FRAC_BITS) + 2 >= RESAMPLE_EXTRA_LEN) {
    /* TODO: Error recover */
    fprintf(stderr, "Too large src_freq / dst_freq ratio: %d -> %d\n",
	    src_freq, dst_freq);
    return -1;
  }
  rs->incr_min = rs->incr - 1;
  rs->incr_max = rs->incr + 1;
  rs->input_count = rs->output_count = 0;

  n = PRECALC_LOOP_COUNT(0, (n<<RESAMPLE_FRAC_BITS), rs->incr);
  rs->outbuf_samples = n;
  rs->outbuf = (short *)malloc(nch * n * sizeof(short));
  return 0;
}

int
resample_change_rate(Resampler *rs, int src_freq, int dst_freq)
{
  int n;

  rs->f_incr = (double)src_freq / (double)dst_freq;
  rs->incr = (int)(rs->f_incr * (1<<RESAMPLE_FRAC_BITS) + 0.5);
  rs->precalc_loop_gard = dst_freq / src_freq + 1;
  if ((rs->incr>>RESAMPLE_FRAC_BITS) + 2 >= RESAMPLE_EXTRA_LEN) {
    /* TODO: Error recover */
    fprintf(stderr, "Too large src_freq / dst_freq ratio: %g -> %g\n",
	    src_freq, dst_freq);
    return -1;
  }
  rs->src_freq = src_freq;
  rs->dst_freq = dst_freq;
  rs->incr_min = rs->incr - 1;
  rs->incr_max = rs->incr + 1;
  rs->input_count = rs->output_count = 0;

  n = rs->inbuf_max + RESAMPLE_EXTRA_LEN;
  n = PRECALC_LOOP_COUNT(0, (n<<RESAMPLE_FRAC_BITS), rs->incr);
  rs->outbuf_samples = n;
  rs->outbuf = (short *)realloc(rs->outbuf, rs->nch * n * sizeof(short));
  return 0;
}


void
resample_free(Resampler *rs)
{
  free(rs->inbuf);
  free(rs->outbuf);
}

static void
resample_linear_convert_1(const short *src, short *dst, int n,
			  int incr, int ofs)
{
  int i;

  for (i = 0; i < n; i++) {
    int v1, v2, j;
    j = (ofs >> RESAMPLE_FRAC_BITS);
    v1 = src[j    ];
    v2 = src[j + 1];
    dst[i] = (short)(v1 + (((v2 - v1) * (ofs & RESAMPLE_FRAC_MASK)) >>
			   RESAMPLE_FRAC_BITS));

    ofs += incr;
  }
}

//static const short *src_debug_end;
static void
resample_linear_convert_2(const short *src, short *dst, int n,
			  int incr, int ofs)
{
  int i;

  for (i = 0; i < n; i++) {
    int l1, r1, l2, r2, j;

    j = (ofs >> RESAMPLE_FRAC_BITS) * 2;
    l1 = src[j    ];
    r1 = src[j + 1];
    l2 = src[j + 2];
    r2 = src[j + 3];

    //if (&src[j + 3] >= src_debug_end) {
    //puts("#### OVERFLOW\n");
    //break;
    //}


    dst[i*2+0] = (short)(l1 + (((l2 - l1) * (ofs & RESAMPLE_FRAC_MASK)) >>
			       RESAMPLE_FRAC_BITS));
    dst[i*2+1] = (short)(r1 + (((r2 - r1) * (ofs & RESAMPLE_FRAC_MASK)) >>
			       RESAMPLE_FRAC_BITS));
    ofs += incr;
  }
}

static void
resample_linear_convert_n(const short *src, short *dst, int n,
			  int incr, int ofs, int nch)
{
  int i;

  for (i = 0; i < n; i++) {
    int j, k, ch;

    j = (ofs >> RESAMPLE_FRAC_BITS) * nch;
    k = i * nch;
    for (ch = 0; ch < nch; ch++) {
      int v1, v2;
      v1 = src[j + ch];
      v2 = src[j + ch + nch];
      dst[k + ch] = (short)(v1 + (((v2 - v1) * (ofs & RESAMPLE_FRAC_MASK)) >>
				  RESAMPLE_FRAC_BITS));
    }
    ofs += incr;
  }
}

int
resample_convert_size(Resampler *rs, int in_samples)
{
  int nch;
  int inbuf_len;

  if (in_samples > rs->inbuf_max) {
    fprintf(stderr, "Too large input data: %d (max=%d)\n", in_samples, rs->inbuf_max);
    return -1;
  }
  nch = rs->nch;
  inbuf_len = rs->inbuf_len;

  if (in_samples > 0)
    inbuf_len += in_samples;

  return PRECALC_LOOP_COUNT(rs->ofs,
			    inbuf_len << RESAMPLE_FRAC_BITS,
			    rs->incr) - 1; /* -1 for
					    * src[(ofs>>RESAMPLE_FRAC_BITS)+1];
					    */
}

static int
resample_check_size(Resampler *rs, int in_samples)
{
  int n;

  /* check */
  n = in_samples + RESAMPLE_EXTRA_LEN;
  n <<= RESAMPLE_FRAC_BITS;
  n >>= RESAMPLE_FRAC_BITS;
  if (n != in_samples + RESAMPLE_EXTRA_LEN) {
    fprintf(stderr, "Too large input size: %d\n", in_samples);
    return -1;
  }

  if (rs->inbuf_size < n) {
    rs->inbuf_size = n * 2;
    rs->inbuf = (short *)realloc(rs->inbuf,
				 rs->nch * rs->inbuf_size * sizeof(short));
  }

  n = PRECALC_LOOP_COUNT(0, (n<<RESAMPLE_FRAC_BITS), rs->incr);
  rs->outbuf_samples = n;

  if (rs->outbuf_samples < n) {
    rs->outbuf_samples = n * 2;
    rs->outbuf = (short *)realloc(rs->outbuf, rs->nch * rs->outbuf_samples * sizeof(short));
  }
  return n;
}

short *
resample_linear_convert(Resampler *rs,
			const short *inbuf, int in_samples,
			int *out_samples)
{
  int i, advanced, n, nch;

  if (resample_check_size(rs, in_samples) == -1) {
    *out_samples = 0;
    fprintf(stderr, "Too large input data: %d (max=%d)\n",
	    in_samples, rs->inbuf_max);
    return NULL;
  }

  nch = rs->nch;

  if (in_samples > 0) {
    memcpy(rs->inbuf + nch * rs->inbuf_len, inbuf,
	   nch * in_samples * sizeof(short));
    rs->inbuf_len += in_samples;
    rs->input_count += in_samples;
  }

  n = PRECALC_LOOP_COUNT(rs->ofs,
			 rs->inbuf_len << RESAMPLE_FRAC_BITS,
			 rs->incr);
  n -= rs->precalc_loop_gard;

  switch (nch) {
  case 1:
    resample_linear_convert_1(rs->inbuf, rs->outbuf, n, rs->incr, rs->ofs);
    break;
  case 2:
    resample_linear_convert_2(rs->inbuf, rs->outbuf, n, rs->incr, rs->ofs);
    break;
  default:
    resample_linear_convert_n(rs->inbuf, rs->outbuf, n, rs->incr, rs->ofs,
			      nch);
    break;
  }
  *out_samples = n;
  rs->output_count += n;
  rs->ofs += n * rs->incr;

  advanced = (rs->ofs >> RESAMPLE_FRAC_BITS);
  n = rs->inbuf_len = rs->inbuf_len - advanced;
  advanced *= nch;
  n *= nch;
  for (i = 0; i < n; i++)
    rs->inbuf[i] = rs->inbuf[advanced + i];
  rs->ofs &= RESAMPLE_FRAC_MASK;

  return rs->outbuf;
}

void
resample_adjust_incr(Resampler *rs)
{
  int target_count;

  target_count = (int)(rs->input_count / rs->f_incr + 0.5);

  //fprintf(stderr, "target=%d, output=%d\n", target_count, rs->output_count);

  if (target_count < rs->output_count)
    rs->incr++;
  else if (target_count > rs->output_count)
    rs->incr--;
  if (rs->incr < rs->incr_min)
    rs->incr = rs->incr_min;
  else if (rs->incr > rs->incr_max)
    rs->incr = rs->incr_max;

  /*fprintf(stderr, "## %f, %d\n",
	  rs->f_incr * (1<<RESAMPLE_FRAC_BITS),
	  rs->incr);*/
  /*usleep(1);*/

  if (rs->input_count > 0x3fffffff || rs->output_count > 0x3fffffff)
    rs->input_count = rs->output_count = 0;
}


#define swap2(x, j, k)	\
{			\
    double tmp;		\
    tmp = x[j];		\
    x[j] = -x[k];	\
    x[k] = tmp;		\
}

#define INBUF_SAMPLES (ALIASING_CONVOLUTION_SIZE * 2)


static void
make_ifft_table(double* itrig, int* b, int n)
{
  unsigned long bitmask;
  int i, j, k;

  /* check n */
  for (i = n; !(i & 1); i >>= 1)
    ;
  if (i != 1) {
    fprintf(stderr, "Invalid fft data size: %d\n", n);
    exit(1);
  }

  /* make bitrev table */
  for (i = 0; i < n; i++)
    b[i] = 0;
  bitmask = n / 2;
  for (i = 1; i < n; i <<= 1, bitmask >>= 1)
    for (j = 0; j < n; j += i * 2)
      for (k = j + i; k < j + i * 2; k++)
	b[k] = (int)((unsigned long)b[k] | bitmask);

  /* make trig table */
  for (i = 0; i < n; i++) {
    double r = (2 * M_PI) * i / (double)n;
    j = i * 2;
    itrig[j  ] = cos(r);
    itrig[j+1] = -sin(r);
  }

  for (i = 0; i < n; i++) {
    if (i < b[i]) {
      double c, d;
      unsigned long i2, j2;

      i2 = i * 2;
      j2 = b[i] * 2;

      c = itrig[i2  ];
      d = itrig[i2+1];

      itrig[i2  ] = itrig[j2  ];
      itrig[i2+1] = itrig[j2+1];

      itrig[j2  ] = c;
      itrig[j2+1] = d;
    }
  }
}

static void
inverse_fft(double* x, double* y, int n)
{
  double* itrig = NULL;
  int* bitrev_table = NULL;
  int i;

  /* initialize */
  itrig = (double *)malloc((n * 2) * sizeof(double));
  bitrev_table = (int *)malloc(n * sizeof(int));
  make_ifft_table(itrig, bitrev_table, n);

  /* main fft proc */
  {
    int j, k;
    int i2, bi;
    int m, ji;
    double c, s;
    double t1x, t1y, t2x, t2y, tx, ty;

    i2 = n / 2;
    for (i = n; i > 1; i = i2, i2 >>= 1) {
      bi = 0;
      for (j = 0; j < n; j += i, bi += 2) {
	c = itrig[bi * 2];
	s = itrig[bi * 2 + 1];

	ji = j + i;
	m = j;
	for (k = j + i2; k < ji; k++, m++) {
	  t1x = x[m];
	  t1y = y[m];
	  tx = x[k];
	  ty = y[k];
	  t2x = c * tx - s * ty;
	  t2y = s * tx + c * ty;
	  x[m] = t1x + t2x;
	  y[m] = t1y + t2y;
	  x[k] = t1x - t2x;
	  y[k] = t1y - t2y;
	}
      }
    }
  }

  /* move data */
  for (i = 0; i < n; i++) {
    if (i < bitrev_table[i]) {
      double tmp;

      tmp = x[i];
      x[i] = x[bitrev_table[i]];
      x[bitrev_table[i]] = tmp;
      tmp = y[i];
      y[i] = y[bitrev_table[i]];
      y[bitrev_table[i]] = tmp;
    }
  }

  /* free */
  free(itrig);
  free(bitrev_table);
}

static void
mkconvolution(double* h)
{
  double hx[ALIASING_CONVOLUTION_SIZE];
  double hy[ALIASING_CONVOLUTION_SIZE];
  int i;
  double t;

  for (i = 0; i <= ALIASING_CONVOLUTION_SIZE/2; i++)
    hx[i] = h[i];
  for (i = 1; i < ALIASING_CONVOLUTION_SIZE/2; i++)
    hx[ALIASING_CONVOLUTION_SIZE - i] = hx[i];
  for (i = 0; i < ALIASING_CONVOLUTION_SIZE; i++)
    hy[i] = 0.0;

  inverse_fft(hx, hy, ALIASING_CONVOLUTION_SIZE);

  for (i = 0; i < ALIASING_CONVOLUTION_SIZE/2; i++) {
    double tmp;

    tmp = hx[i];
    hx[i] = hx[i + ALIASING_CONVOLUTION_SIZE/2];
    hx[i + ALIASING_CONVOLUTION_SIZE/2] = tmp;
  }

  /* Window */
  t = -M_PI;
  for (i = 0; i <= ALIASING_CONVOLUTION_SIZE/2; i++) {
    hx[i] *= (0.50 + 0.50 * cos(t)) / ALIASING_CONVOLUTION_SIZE;
    t += 2.0 * M_PI / ALIASING_CONVOLUTION_SIZE;
  }

  memcpy(h, hx, (ALIASING_CONVOLUTION_SIZE/2 + 1) * sizeof(double));
}

int
anti_aliasing_init(AntiAliasing *aa, int nch, int bufsiz,
		   int src_freq, int dst_freq)
{
  int i, n;
  int lowpath;

  if (!bufsiz)
    bufsiz = RESAMPLE_DEFAULT_BUFSIZ;

  memset(aa, 0, sizeof(*aa));
  aa->nch = nch;

  if (src_freq <= dst_freq)
    return 0;

  n = INBUF_SAMPLES * nch * sizeof(short);
  aa->inbuf = (short *)malloc(n);
  memset(aa->inbuf, 0, n);
  aa->inbuf_len = ALIASING_CONVOLUTION_SIZE/2;

  if (bufsiz < ALIASING_CONVOLUTION_SIZE)
    bufsiz = ALIASING_CONVOLUTION_SIZE;
  aa->outbuf_size = bufsiz;
  n = aa->outbuf_size * nch * sizeof(short);
  aa->outbuf = (short *)malloc(n);
  memset(aa->outbuf, 0, n);

  lowpath = (int)((ALIASING_CONVOLUTION_SIZE/2) * ((double)dst_freq/src_freq));
  for (i = 0; i < lowpath; i++)
    aa->fh[i] = 1.0;
  for (; i < ALIASING_CONVOLUTION_SIZE/2; i++)
    aa->fh[i] = 0.0;
  aa->fh[ALIASING_CONVOLUTION_SIZE/2] = aa->fh[ALIASING_CONVOLUTION_SIZE/2-1];

  mkconvolution(aa->fh);
  for (i = 0; i < ALIASING_CONVOLUTION_SIZE/2+1; i++)
    aa->ih[i] = (int)(aa->fh[i] * 65536.0 + 0.5);
  return 0;
}

void
anti_aliasing_free(AntiAliasing *aa)
{
  if (aa->inbuf)
    free(aa->inbuf);
  if (aa->outbuf)
    free(aa->outbuf);
}

static const short *
anti_aliasing_fill(AntiAliasing *aa, const short *inbuf, int *in_out_samples)
{
  int n, nch;

  nch = aa->nch;
  n = INBUF_SAMPLES - aa->inbuf_len;

  //fprintf(stderr, "### n=%d\n", n);

  /*DEBUG*/
  if (aa->inbuf_len < ALIASING_CONVOLUTION_SIZE/2 || aa->inbuf_len > ALIASING_CONVOLUTION_SIZE)
    fprintf(stderr, "### Oops, There, the bug exists: n=%d\n", n);

  if (inbuf == NULL) {
    /* EOF */
    n = ALIASING_CONVOLUTION_SIZE/2;
    memset(aa->inbuf + aa->inbuf_len * nch,
	   0,
	   n * nch * sizeof(short));
    aa->inbuf_len += n;
    *in_out_samples = 0;
    return NULL;
  }


  if (n > *in_out_samples)
    n = *in_out_samples;
  memcpy(aa->inbuf + aa->inbuf_len * nch,
	 inbuf,
	 n * nch * sizeof(short));
  aa->inbuf_len += n;
  (*in_out_samples) -= n;
  return inbuf + n * nch;
}

static void
anti_aliasing_slide(AntiAliasing *aa, int n_out,
		    const short *inbuf, int in_samples)
{
  int n, nch;

  nch = aa->nch;
  if (in_samples >= ALIASING_CONVOLUTION_SIZE) {
    memcpy(aa->inbuf,
	   inbuf + (in_samples - ALIASING_CONVOLUTION_SIZE) * nch,
	   ALIASING_CONVOLUTION_SIZE * nch * sizeof(short));
  } else if (in_samples > 0) {
    n = ALIASING_CONVOLUTION_SIZE - in_samples;
    memmove(aa->inbuf,
	    aa->inbuf + (INBUF_SAMPLES - n) * nch,
	    n * nch * sizeof(short));
    memmove(aa->inbuf + n * nch,
	    inbuf,
	    in_samples * nch * sizeof(short));
  } else {
    memmove(aa->inbuf,
	    aa->inbuf + n_out * nch,
	    (aa->inbuf_len - n_out) * nch * sizeof(short));
  }
  aa->inbuf_len = ALIASING_CONVOLUTION_SIZE;
}

static void
anti_aliasing_convert_1(const short *src, short *dst, int n, double *h)
{
  int i;

  for (i = 0; i < n; i++, src++) {
    double v;
    int j, k;

    v = h[0] * src[0] +
      h[ALIASING_CONVOLUTION_SIZE/2] * src[ALIASING_CONVOLUTION_SIZE/2];
    for(j = 1, k = ALIASING_CONVOLUTION_SIZE - 1;
	j < ALIASING_CONVOLUTION_SIZE/2; j++, k--)
      v += h[j] * (src[j] + src[k]);
#if 0
    s = v;
    if (s < -32768)
      s = -32768;
    else if (s > 32767)
      s = 32767;
    *dst++ = s;
#else
    *dst++ = (short)v;
#endif
  }
}

static void
anti_aliasing_convert_fast_1(const short *src, short *dst, int n, int *h)
{
  int i;

  for (i = 0; i < n; i++, src++) {
    int v, j, k;

    v = (h[0] * src[0] +
	 h[ALIASING_CONVOLUTION_SIZE/2] * src[ALIASING_CONVOLUTION_SIZE/2])
	   >> 16;
    for(j = 1, k = ALIASING_CONVOLUTION_SIZE - 1;
	j < ALIASING_CONVOLUTION_SIZE/2; j++, k--)
      v += (h[j] * (src[j] + src[k])) >> 16;
    *dst++ = v;
  }
}

static void
anti_aliasing_convert_2(const short *src, short *dst, int n, double *h)
{
  int i;

  for (i = 0; i < n; i++, src += 2) {
    double v1, v2;
    int j, k;

    v1 = h[0] * src[0] +
      h[ALIASING_CONVOLUTION_SIZE/2] * src[ALIASING_CONVOLUTION_SIZE];
    v2 = h[0] * src[1] +
      h[ALIASING_CONVOLUTION_SIZE/2] * src[ALIASING_CONVOLUTION_SIZE+1];

    for(j = 1, k = ALIASING_CONVOLUTION_SIZE - 1;
	j < ALIASING_CONVOLUTION_SIZE/2; j++, k--) {
      int jj = j * 2;
      int kk = k * 2;
      v1 += h[j] * (src[jj] + src[kk]);
      v2 += h[j] * (src[jj+1] + src[kk+1]);
    }
#if 0
    s = v1;
    if (s < -32768)
      s = -32768;
    else if (s > 32767)
      s = 32767;
    *dst++ = s;

    s = v2;
    if (s < -32768)
      s = -32768;
    else if (s > 32767)
      s = 32767;
    *dst++ = s;
#else
    *dst++ = (short)v1;
    *dst++ = (short)v2;
#endif
  }
}

static void
anti_aliasing_convert_fast_2(const short *src, short *dst, int n, int *h)
{
  int i;

  for (i = 0; i < n; i++, src += 2) {
    int v1, v2;
    int j, k;

    v1 = (h[0] * src[0] +
      h[ALIASING_CONVOLUTION_SIZE/2] * src[ALIASING_CONVOLUTION_SIZE]) >> 16;
    v2 = (h[0] * src[1] +
      h[ALIASING_CONVOLUTION_SIZE/2] * src[ALIASING_CONVOLUTION_SIZE+1]) >> 16;

    for(j = 1, k = ALIASING_CONVOLUTION_SIZE - 1;
	j < ALIASING_CONVOLUTION_SIZE/2; j++, k--) {
      int jj = j * 2;
      int kk = k * 2;
      v1 += (h[j] * (src[jj] + src[kk])) >> 16;
      v2 += (h[j] * (src[jj+1] + src[kk+1])) >> 16;
    }
    *dst++ = v1;
    *dst++ = v2;
  }
}

static void
anti_aliasing_convert_n(const short *src, short *dst, int n, double *h,
			int nch)
{
  int i;

  for (i = 0; i < n; i++, src += nch) {
    double v[ALIASING_MAX_CHANNELS];
    int j, k;
    int ch;

    for (ch = 0; ch < nch; ch++)
      v[ch] = h[0] * src[ch] +
	h[ALIASING_CONVOLUTION_SIZE/2] *
	src[ALIASING_CONVOLUTION_SIZE/2*nch + ch];

    for(j = 1, k = ALIASING_CONVOLUTION_SIZE - 1;
	j < ALIASING_CONVOLUTION_SIZE/2; j++, k--) {
      int jj = j * nch;
      int kk = k * nch;
      int ch;
      double hj;

      hj = h[j];
      for (ch = 0; ch < nch; ch++)
	v[ch] += hj * (src[jj+ch] + src[kk+ch]);
    }
    for (ch = 0; ch < nch; ch++) {
#if 0
      s = v[ch];
      if (s < -32768)
	s = -32768;
      else if (s > 32767)
	s = 32767;
      *dst++ = s;
#else
      *dst++ = (short)v[ch];
#endif
    }
  }
}

static void
anti_aliasing_convert_fast_n(const short *src, short *dst, int n, int *h,
			     int nch)
{
  int i;

  for (i = 0; i < n; i++, src += nch) {
    int v[ALIASING_MAX_CHANNELS];
    int j, k;
    int ch;

    for (ch = 0; ch < nch; ch++)
      v[ch] = (h[0] * src[ch] +
	h[ALIASING_CONVOLUTION_SIZE/2] *
	src[ALIASING_CONVOLUTION_SIZE/2*nch + ch]) >> 16;

    for(j = 1, k = ALIASING_CONVOLUTION_SIZE - 1;
	j < ALIASING_CONVOLUTION_SIZE/2; j++, k--) {
      int jj = j * nch;
      int kk = k * nch;
      int ch;
      int hj;

      hj = h[j];
      for (ch = 0; ch < nch; ch++)
	v[ch] += (hj * (src[jj+ch] + src[kk+ch])) >> 16;
    }
    for (ch = 0; ch < nch; ch++)
      *dst++ = v[ch];
  }
}

short *
anti_aliasing_convert(AntiAliasing *aa,
		      const short *inbuf, int in_samples,
		      int *out_samples)
{
  short *obuf;
  int n, nch, out_size;

  if (aa->inbuf == NULL) {
    *out_samples = in_samples;
    return (short *)inbuf;
  }

  inbuf = anti_aliasing_fill(aa, inbuf, &in_samples);
  if (aa->inbuf_len <= ALIASING_CONVOLUTION_SIZE) {
    *out_samples = 0;
    return aa->outbuf;
  }

  n = aa->inbuf_len - ALIASING_CONVOLUTION_SIZE;
  nch = aa->nch;

  out_size = (n + in_samples) * nch;
  if (out_size > aa->outbuf_size) {
    aa->outbuf_size = 2 * out_size;
    aa->outbuf = (short *)realloc(aa->outbuf,
				  aa->outbuf_size * nch * sizeof(short));
  }
  obuf = aa->outbuf;

  switch (nch) {
  case 1:
    anti_aliasing_convert_1(aa->inbuf, obuf, n, aa->fh);
    break;
  case 2:
    anti_aliasing_convert_2(aa->inbuf, obuf, n, aa->fh);
    break;
  default:
    anti_aliasing_convert_n(aa->inbuf, obuf, n, aa->fh, nch);
    break;
  }
  obuf += n * nch;
  *out_samples = n;

  if (in_samples == 0) {
    anti_aliasing_slide(aa, *out_samples, NULL, 0);
    return aa->outbuf;
  }

  switch (nch) {
  case 1:
    anti_aliasing_convert_1(inbuf - nch * ALIASING_CONVOLUTION_SIZE,
			    obuf, in_samples, aa->fh);
    break;
  case 2:
    anti_aliasing_convert_2(inbuf - nch * ALIASING_CONVOLUTION_SIZE,
			    obuf, in_samples, aa->fh);
    break;
  default:
    anti_aliasing_convert_n(inbuf - nch * ALIASING_CONVOLUTION_SIZE,
			    obuf, in_samples, aa->fh, nch);
    break;
  }

  (*out_samples) += in_samples;

  anti_aliasing_slide(aa, *out_samples, inbuf, in_samples);
  return aa->outbuf;
}

short *
anti_aliasing_convert_fast(AntiAliasing *aa,
			   const short *inbuf, int in_samples,
			   int *out_samples)
{
  short *obuf;
  int n, nch;
  int out_size;

  if (aa->inbuf == NULL) {
    *out_samples = in_samples;
    return (short *)inbuf;
  }

  inbuf = anti_aliasing_fill(aa, inbuf, &in_samples);
  if (aa->inbuf_len <= ALIASING_CONVOLUTION_SIZE) {
    *out_samples = 0;
    return aa->outbuf;
  }

  n = aa->inbuf_len - ALIASING_CONVOLUTION_SIZE;
  nch = aa->nch;

  out_size = (n + in_samples) * nch;
  if (out_size > aa->outbuf_size) {
    aa->outbuf_size = 2 * out_size;
    aa->outbuf = (short *)realloc(aa->outbuf,
				  aa->outbuf_size * nch * sizeof(short));
  }
  obuf = aa->outbuf;

  switch (nch) {
  case 1:
    anti_aliasing_convert_fast_1(aa->inbuf, obuf, n, aa->ih);
    break;
  case 2:
    anti_aliasing_convert_fast_2(aa->inbuf, obuf, n, aa->ih);
    break;
  default:
    anti_aliasing_convert_fast_n(aa->inbuf, obuf, n, aa->ih, nch);
    break;
  }
  obuf += n * nch;
  *out_samples = n;

  if (in_samples == 0) {
    anti_aliasing_slide(aa, *out_samples, NULL, 0);
    return aa->outbuf;
  }

  switch (nch) {
  case 1:
    anti_aliasing_convert_fast_1(inbuf - nch * ALIASING_CONVOLUTION_SIZE,
				 obuf, in_samples, aa->ih);
    break;
  case 2:
    anti_aliasing_convert_fast_2(inbuf - nch * ALIASING_CONVOLUTION_SIZE,
				 obuf, in_samples, aa->ih);
    break;
  default:
    anti_aliasing_convert_fast_n(inbuf - nch * ALIASING_CONVOLUTION_SIZE,
				 obuf, in_samples, aa->ih, nch);
    break;
  }

  (*out_samples) += in_samples;

  anti_aliasing_slide(aa, *out_samples, inbuf, in_samples);
  return aa->outbuf;
}

int
anti_aliasing_resample_init(AntiAliasingResample *aar,
			    int nch, int src_freq, int dst_freq)
{
  memset(aar, 0, sizeof(*aar));
  if (src_freq != dst_freq)
    resample_init(&aar->rs, nch, 0, src_freq, dst_freq);
  if (src_freq > dst_freq)
    anti_aliasing_init(&aar->aa, nch, 0,  src_freq, dst_freq);
  aar->nch = nch;
  return 0;
}

void
anti_aliasing_resample_free(AntiAliasingResample *aar)
{
  if (aar->rs.inbuf)
    resample_free(&aar->rs);
  if (aar->aa.inbuf)
    anti_aliasing_free(&aar->aa);
}

char *
anti_aliasing_resample(AntiAliasingResample *aar,
		       const char *data, int in_size, int *out_size)
{
  short *resample_data;
  int    resample_size;
  char *ret_data;
  int   ret_size;
  int   block_align;

  block_align = aar->nch * sizeof(short);

  if (aar->aa.inbuf) {
    resample_data =
      anti_aliasing_convert_fast(&aar->aa, (short *)data,
				 in_size / block_align, &resample_size);
    //fprintf(stderr, "## AA: %d -> %d\n", in_size / block_align, resample_size);

  } else {
    resample_data = (short *)data;
    resample_size = in_size / (aar->nch * sizeof(short));
  }

  if (aar->rs.inbuf) {
    ret_data = (char *)resample_linear_convert(&aar->rs,
					       resample_data, resample_size,
					       &ret_size);
    resample_adjust_incr(&aar->rs);
  } else {
    ret_data = (char *)resample_data;
    ret_size = resample_size;
  }

  *out_size = ret_size * block_align;
  return ret_data;
}



#define DO_UP_RESAMPLING		\
  int i, j;				\
  static int cc = 0;			\
  j = 0;				\
  for (i = 0; i < nsamples; i++) {	\
    if (cc == 0)			\
      cc = ratio;			\
    else				\
      pcm[j++] = pcm[i];		\
    cc--;				\
  }					\
  return j

static int
simple_up_resampling_8(uint8_t *pcm, int nsamples, int ratio)
{
  DO_UP_RESAMPLING;
}

static int
simple_up_resampling_16(uint16_t *pcm, int nsamples, int ratio)
{
  DO_UP_RESAMPLING;
}

static int
simple_up_resampling_32(uint32_t *pcm, int nsamples, int ratio)
{
  DO_UP_RESAMPLING;
}

int
simple_up_resampling(void *pcm, int bits, int nch, int size, int ratio)
{
  if (bits == 8) {
    if (nch == 1) {
      return simple_up_resampling_8((uint8_t *)pcm, size, ratio);
    } else {
      return simple_up_resampling_16((uint16_t *)pcm, size>>1, ratio)<<1;
    }
  } else {
    if (nch == 1) {
      return simple_up_resampling_16((uint16_t *)pcm, size>>1, ratio)<<1;
    } else {
      return simple_up_resampling_32((uint32_t *)pcm, size>>2, ratio)<<2;
    }
  }
}


#if 0
/* LMediaQueueFilterFunction: AntiAliasingResample *arg */
/* Note: 32*dst_freq/src_freq samples are remained at the end of stream  */
LMediaPacket *
aar_convert_filter(LMediaPacket *packet, void *arg)
{
  LMediaPacket *new_packet;
  char *data;
  int   size;

  if (lmq_is_eos(packet))
    return packet;
  if (packet->type != lmm_TAG_AUDS)
    return packet;

  data = anti_aliasing_resample((AntiAliasingResample *)arg,
				LMQ_PACKET_DATA(packet),
				LMQ_PACKET_SIZE(packet),
				&size);

  new_packet = lmq_alloc_packet(size);
  memcpy(new_packet, packet, LMQ_SIZEOF_PACKET);
  lmq_release(packet);

  memcpy(LMQ_PACKET_DATA(new_packet), data, size);
  new_packet->size = size;

  return new_packet;
}
#endif


#if 0
#include "izmd/izmd.h"
int
main(int argc, char **argv)
{
  int src_freq = 44100;
  int dst_freq = 8000;
  char buff[BUFSIZ];
  AntiAliasingResample aar;
  int n;
  char *data;
  int   size;

  //izmd_trace();

  anti_aliasing_resample_init(&aar, 2, src_freq, dst_freq);

  while ((n = fread(buff, 1, sizeof(buff), stdin)) > 0) {

    data = anti_aliasing_resample(&aar, buff, n, &size);
    fwrite(data, 1, size, stdout);
  }

  data = anti_aliasing_resample(&aar, NULL, 0, &size);
  fwrite(data, 1, size, stdout);


  anti_aliasing_resample_free(&aar);

  //izmd_untrace();

  return 0;
}
#endif
